/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "alembic/alembic.pmh"

#define FE_SAA_DEBUG			FALSE
#define FE_SAA_VERBOSE			FALSE

//* TODO properly use the per-vertex normals/uvs (need better RG vertex support)
#define FE_SAA_UNIQUE_POINTS	FALSE

namespace fe
{

SurfaceAccessibleAbc::FileContext::FileContext(void)
{
}

SurfaceAccessibleAbc::FileContext::~FileContext(void)
{
	m_archive.reset();
}

BWORD SurfaceAccessibleAbc::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load file \"%s\"\n",a_filename.c_str());
#endif

	//* TODO option to pick mesh, curve, or joint

	if(a_spSettings.isValid())
	{
		m_frame=a_spSettings->catalog<Real>("frame");
		m_spFilter=a_spSettings->catalog< sp<Component> >("filter");

		const String options=a_spSettings->catalog<String>("options");

#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load options \"%s\"\n",options.c_str());
#endif

		if(!options.empty())
		{
			setOptions(options);
		}
	}
	else
	{
		m_frame=0.0;
		m_spFilter=NULL;
	}

	m_fps= -1.0;

	m_minStep= -1.0;

	m_refMode=(m_optionMap.find("ref")!=m_optionMap.end());

	if(m_optionMap.find("fps")!=m_optionMap.end())
	{
		m_fps=atof(m_optionMap["fps"].c_str());
	}

	m_pathArray.clear();
	if(m_optionMap.find("paths")!=m_optionMap.end())
	{
		String pathList=m_optionMap["paths"];
		String path;

		while(!(path=pathList.parse()).empty())
		{
			if(path.c_str()[0]!='/')
			{
				path="/"+path;
			}

			m_pathArray.push_back(path);
		}
	}

	m_matchArray.clear();
	if(m_optionMap.find("match")!=m_optionMap.end())
	{
		String matchList=m_optionMap["match"];
		String match;

		while(!(match=matchList.parse()).empty())
		{
			m_matchArray.push_back(match);
		}
	}

	if(m_filename!=a_filename)
	{
		m_spContext=NULL;
		m_filename=a_filename;
	}

	if(m_spContext.isValid() && !m_spContext->m_archive)
	{
		m_spContext=NULL;
	}

#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load frame %.6G\n",m_frame);
#endif

	if(m_spContext.isNull())
	{

		setContext(a_filename);
	}

	if(m_spContext.isNull() || !m_spContext->m_archive.valid())
	{
		feLog("SurfaceAccessibleAbc::load failed to read \"%s\"\n",
				a_filename.c_str());

		return FALSE;
	}

	Real minTimePerSample= -1.0;

	const I32 samplerCount=m_spContext->m_archive.getNumTimeSamplings();
	for(I32 samplerIndex=0;samplerIndex<samplerCount;samplerIndex++)
	{
		Alembic::AbcGeom::TimeSamplingPtr sampler=
				m_spContext->m_archive.getTimeSampling(samplerIndex);

		Alembic::AbcGeom::TimeSamplingType samplerType=
				sampler->getTimeSamplingType();

		const BWORD isUniform=samplerType.isUniform();
		const Real timePerCycle=samplerType.getTimePerCycle();

#if FE_SAA_DEBUG
		const I32 timeCount=sampler->getNumStoredTimes();
		const BWORD isCyclic=samplerType.isCyclic();
		const BWORD isAcyclic=samplerType.isAcyclic();
		const U32 samplesPerCycle=samplerType.getNumSamplesPerCycle();
		const Real fps=Real(1)/timePerCycle;

		feLog("sampler %d/%d times %d uni %d cyc %d acyc %d"
				" SPC %d TPC %.6G FPS %.6G\n",
				samplerIndex,samplerCount,timeCount,
				isUniform,isCyclic,isAcyclic,
				samplesPerCycle,timePerCycle,fps);
#endif

		if(!isUniform)
		{
			feLog("SurfaceAccessibleAbc::load"
					" non-uniform sampling detected\n");
		}

		if(timePerCycle>1e-3 &&
				(minTimePerSample<0.0 || minTimePerSample>timePerCycle))
		{
			minTimePerSample=timePerCycle;
		}
	}

	m_mode=e_scan;
	m_xformCount=0;
	m_meshCount=0;
	m_curveCount=0;

	Alembic::AbcGeom::IObject iObj=m_spContext->m_archive.getTop();

	visit(iObj);

#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load"
			" scan xformCount %d meshCount %d curveCount %d"
			" minStep %.6G minTimePerSample %.6G\n",
			m_xformCount,m_meshCount,m_curveCount,
			m_minStep,minTimePerSample);
#endif

	//* auto-detected frames per second
	if(m_fps<0.0)
	{
		if(m_minStep>0.0)
		{
			m_fps=Real(1)/m_minStep;
		}
		else if(minTimePerSample>0.0)
		{
			m_fps=Real(1)/minTimePerSample;
		}
		else
		{
			m_fps=Real(24.0);
		}

#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load presuming %.6G fps\n",m_fps);
#endif
	}

	m_firstFaceMap.clear();

	//* NOTE trying to auto-detect intended content
	if(m_meshCount)
	{
		if(m_curveCount)
		{
			feLog("SurfaceAccessibleAbc::load ignoring curves in mesh mode\n");
		}

		m_mode=e_mesh;

		startMeshes();

#if !FE_SAA_UNIQUE_POINTS
		m_spOutputUV=accessor(e_vertex,e_uv,
				SurfaceAccessibleI::e_createMissing);
#endif

		visit(iObj);
		completeMeshes();
	}
	else if(m_curveCount)
	{
		m_mode=e_curve;

		startCurves();
		visit(iObj);
		completeCurves();
	}
	else if(m_xformCount>1)
	{
		m_mode=e_joint;

		startJoints();
		visit(iObj);
		completeJoints();
	}

//	SurfaceAccessibleRecord::save("abc.rg");

#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load done\n");
#endif

	return TRUE;
}

BWORD SurfaceAccessibleAbc::setContext(String a_filename)
{
	sp<Catalog> spMasterCatalog=registry()->master()->catalog();
	const String key="SurfaceAccessibleAbc:"+a_filename;

	m_spContext=spMasterCatalog->catalog< hp<Component> >(key);

	if(m_spContext.isValid())
	{
		return TRUE;
	}

#if FE_SFB_DEBUG
	feLog("SurfaceAccessibleAbc::setContext"
			" creating new FileContext\n");
#endif

	m_spContext=new FileContext();
	spMasterCatalog->catalog< hp<Component> >(key)=m_spContext;

	if(m_spContext.isNull())
	{
		feLog("SurfaceAccessibleAbc::setContext"
				" unable to create FileContext\n");
		return FALSE;
	}

#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::setContext opening archive\n");
#endif

	Alembic::AbcGeom::ErrorHandler::Policy policy=
//			Alembic::AbcGeom::ErrorHandler::kQuietNoopPolicy:
			Alembic::AbcGeom::ErrorHandler::kNoisyNoopPolicy;

#if ALEMBIC_LIBRARY_VERSION < 10500
	Alembic::AbcGeom::IArchive* pArchive=new Alembic::AbcGeom::IArchive(
			Alembic::AbcCoreHDF5::ReadArchive(),a_filename.c_str(),policy);

	//* TODO verify that this is properly ref-counted underneath
	m_spContext->m_archive= *pArchive;
	delete pArchive;
#else
	Alembic::AbcCoreFactory::IFactory factory;
	factory.setPolicy(policy);
	m_spContext->m_archive=factory.getArchive(a_filename.c_str());
#endif

	if(!m_spContext->m_archive.valid())
	{
		m_spContext=NULL;
		return FALSE;
	}

#if FE_SAA_DEBUG
	std::string appName;
	std::string libraryVersionString;
	Alembic::Util::uint32_t libraryVersion;
	std::string whenWritten;
	std::string userDescription;

	Alembic::Abc::GetArchiveInfo(m_spContext->m_archive,appName,
			libraryVersionString,libraryVersion,whenWritten,userDescription);

	feLog("  written by    \"%s\"\n",appName.c_str());
	feLog("  using Alembic %d \"%s\"\n",
			libraryVersion,libraryVersionString.c_str());
	feLog("  written on    \"%s\"\n",whenWritten.c_str());
	feLog("  description   \"%s\"\n",userDescription.c_str());
#endif

	return TRUE;
}

void SurfaceAccessibleAbc::visit(Alembic::AbcGeom::IObject a_iObj)
{
	const String parentPath=a_iObj.getParent().getFullName().c_str();
	const String path=a_iObj.getFullName().c_str();
	const U32 childCount=a_iObj.getNumChildren();

#if FE_SAA_DEBUG
	if(m_mode!=e_scan)
	{
		feLog("\nobject \"%s\" childCount %d\n",path.c_str(),childCount);
	}
#endif

	setIdentity(m_refWorld);

	Alembic::AbcGeom::ICompoundProperty properties=a_iObj.getProperties();
	String indent="";

	m_pointArray.clear();
	m_pointRefArray.clear();
	m_colorArray.clear();
	m_normalArray.clear();
	m_uvArray.clear();
	m_uvLookupArray.clear();
	m_vertCountArray.clear();
	m_vertIndexArray.clear();

	m_colorScope=Alembic::AbcGeom::kUnknownScope;
	m_uvScope=Alembic::AbcGeom::kUnknownScope;

	BWORD allowed=TRUE;

	const I32 pathCount=m_pathArray.size();
	if(pathCount)
	{
		I32 pathIndex=0;
		for(;pathIndex<pathCount;pathIndex++)
		{
			const String& rPath=m_pathArray[pathIndex];
			if(path==rPath || path.match(rPath+"/.*"))
			{
//				feLog("PATH MATCH \"%s\" vs \"%s\"\n",
//						path.c_str(),rPath.c_str());
				break;
			}
		}
		if(pathIndex>=pathCount)
		{
			allowed=FALSE;
		}
	}

	const I32 matchCount=m_matchArray.size();
	if(allowed && matchCount)
	{
		I32 matchIndex=0;
		for(;matchIndex<matchCount;matchIndex++)
		{
			const String& rMatch=m_matchArray[matchIndex];
			if(path.match(rMatch))
			{
//				feLog("REGEX MATCH \"%s\" vs \"%s\"\n",
//						path.c_str(),rMatch.c_str());
				break;
			}
		}
		if(matchIndex>=matchCount)
		{
			allowed=FALSE;
		}
	}

	if(allowed && m_mode!=e_scan)
	{
		visitProperties(properties,indent,parentPath,path);
	}

	//* TODO add targets

	if(!allowed)
	{
		//* SKIP
	}
	else if(Alembic::AbcGeom::IPolyMesh::matches(a_iObj.getHeader()) ||
			Alembic::AbcGeom::ISubD::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			if(Alembic::AbcGeom::IPolyMesh::matches(a_iObj.getHeader()))
			{
				feLog("MESH\n");
			}
			else
			{
				feLog("SUBDIV\n");
			}
		}
#endif
		if(m_mode==e_scan)
		{
			m_meshCount++;
		}
		if(m_mode==e_mesh)
		{
			const I32 existingPointCount=m_spOutputPoint->count();
			const I32 existingFaceCount=m_spOutputVertices->count();

			m_firstFaceMap[path]=existingFaceCount;

			const I32 addFaceCount=m_vertCountArray.size();

			const Array<SpatialVector>& rPointArray=
					(m_refMode && m_pointRefArray.size())?
					m_pointRefArray: m_pointArray;

#if FE_SAA_UNIQUE_POINTS
			I32 addPointCount=0;
			for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
			{
				addPointCount+=m_vertCountArray[faceIndex];
			}

			m_spOutputPoint->append(addPointCount);

			for(I32 index=0;index<addPointCount;index++)
			{
				const I32 lookupIndex=m_vertIndexArray[index];
				const I32 pointIndex=existingPointCount+index;
				m_spOutputPoint->set(pointIndex,rPointArray[lookupIndex]);
			}
#else
			const I32 addPointCount=rPointArray.size();
			m_spOutputPoint->append(addPointCount);

			for(I32 index=0;index<addPointCount;index++)
			{
				const I32 pointIndex=existingPointCount+index;
				m_spOutputPoint->set(pointIndex,rPointArray[index]);
			}
#endif

#if FE_SAA_DEBUG
			feLog("  %d points %d faces\n",addPointCount,addFaceCount);
#endif

			m_primVerts.resize(addFaceCount);
			for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
			{
				m_primVerts[faceIndex].resize(m_vertCountArray[faceIndex]);
			}

			I32 vertIndex=0;

			const I32 addVertexCount=m_vertIndexArray.size();

			for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
			{
				Array<I32>& rVerts=m_primVerts[faceIndex];
				const U32 vertexCount=rVerts.size();

				for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
				{
					if(vertIndex>=addVertexCount)
					{
						feLog("SurfaceAccessibleAbc::visit"
								" face counts exceeds face indices\n");
						faceIndex=addFaceCount;
						break;
					}
#if FE_SAA_UNIQUE_POINTS
					rVerts[vertexIndex]=
							existingPointCount+(vertIndex++);
#else
					rVerts[vertexIndex]=
							existingPointCount+m_vertIndexArray[vertIndex++];
#endif
				}
			}

			m_spOutputVertices->append(m_primVerts);

			for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
			{
				const I32 primitiveIndex=existingFaceCount+faceIndex;

				m_spOutputName->set(primitiveIndex,path.c_str());
			}

			if(m_colorArray.size())
			{
				vertIndex=0;

				const BWORD constantRate=
						(m_colorScope==Alembic::AbcGeom::kConstantScope);
				const BWORD faceRate=
						(m_colorScope==Alembic::AbcGeom::kUniformScope);

				for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
				{
					const I32 primitiveIndex=existingFaceCount+faceIndex;

					Array<I32>& rVerts=m_primVerts[faceIndex];
					const U32 vertexCount=rVerts.size();
					for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
					{
#if FE_SAA_UNIQUE_POINTS
						const I32 pointIndex=existingPointCount+vertIndex;
#else
						const I32 pointIndex=m_spOutputVertices->integer(
								primitiveIndex,vertexIndex);
#endif
						const I32 lookupIndex=constantRate? 0:
								(faceRate? faceIndex: vertIndex);
						m_spOutputColor->set(pointIndex,
								m_colorArray[lookupIndex]);

						vertIndex++;
					}
				}
			}

			if(m_normalArray.size())
			{
				vertIndex=0;

				for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
				{
					const I32 primitiveIndex=existingFaceCount+faceIndex;

					Array<I32>& rVerts=m_primVerts[faceIndex];
					const U32 vertexCount=rVerts.size();
					for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
					{
#if FE_SAA_UNIQUE_POINTS
						const I32 pointIndex=existingPointCount+vertIndex;
#else
						const I32 pointIndex=m_spOutputVertices->integer(
								primitiveIndex,vertexIndex);
#endif
						m_spOutputNormal->set(pointIndex,
								m_normalArray[vertIndex++]);
					}
				}
			}
			else
			{
				vertIndex=0;

				for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
				{
					const I32 primitiveIndex=existingFaceCount+faceIndex;

					Array<I32>& rVerts=m_primVerts[faceIndex];

					const I32 index0=m_vertIndexArray[vertIndex];
					const I32 index1=m_vertIndexArray[vertIndex+1];
					const I32 index2=m_vertIndexArray[vertIndex+2];

					const SpatialVector point0=rPointArray[index0];
					const SpatialVector point1=rPointArray[index1];
					const SpatialVector point2=rPointArray[index2];

					const SpatialVector norm=
							unitSafe(cross(point2-point0,point1-point0));

					//* NOTE using face normal on points is particularly sloppy

					const U32 vertexCount=rVerts.size();
					for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
					{
#if FE_SAA_UNIQUE_POINTS
						const I32 pointIndex=existingPointCount+vertIndex;
#else
						const I32 pointIndex=m_spOutputVertices->integer(
								primitiveIndex,vertexIndex);
#endif
						m_spOutputNormal->set(pointIndex,norm);

						vertIndex++;
					}
				}
			}

			if(m_uvArray.size())
			{
				vertIndex=0;

				if(m_uvScope!=Alembic::AbcGeom::kFacevaryingScope)
				{
					feLog("SurfaceAccessibleAbc::visit"
							" WARNING: uv data is not kFacevaryingScope\n");
				}

				for(I32 faceIndex=0;faceIndex<addFaceCount;faceIndex++)
				{
					const I32 primitiveIndex=existingFaceCount+faceIndex;

					Array<I32>& rVerts=m_primVerts[faceIndex];
					const U32 vertexCount=rVerts.size();
					for(U32 vertexIndex=0;vertexIndex<vertexCount;vertexIndex++)
					{
						const I32 uvIndex=m_uvLookupArray[vertIndex++];
						const SpatialVector uvValue=m_uvArray[uvIndex];

#if FE_SAA_UNIQUE_POINTS
						const I32 pointIndex=existingPointCount+vertIndex;

						m_spOutputUV->set(pointIndex,uvValue);
#else
//						const I32 pointIndex=m_spOutputVertices->integer(
//								primitiveIndex,vertexIndex);

						m_spOutputUV->set(primitiveIndex,vertexIndex,uvValue);
#endif


#if FALSE
						const SpatialVector uvColor(
								0.125*uvValue[0],0.5*uvValue[1],0.0);
						m_spOutputColor->set(pointIndex,uvColor);
#endif

//						feLog("pr %d %d pt %d lk %d uv %s\n",
//								primitiveIndex,vertexIndex,pointIndex,
//								uvIndex,c_print(uvValue));

					}
				}
			}
		}
	}
	else if(Alembic::AbcGeom::ICurves::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			feLog("CURVES\n");
		}
#endif
		if(m_mode==e_scan)
		{
			m_curveCount++;
		}
		if(m_mode==e_curve)
		{
			const I32 existingPointCount=m_spOutputPoint->count();

			const I32 addPointCount=m_pointArray.size();
			m_spOutputPoint->append(addPointCount);

#if FE_SAA_DEBUG
			feLog("  %d points\n",addPointCount);
#endif

			for(I32 index=0;index<addPointCount;index++)
			{
				const I32 pointIndex=existingPointCount+index;
				m_spOutputPoint->set(pointIndex,m_pointArray[index]);
			}

			const I32 existingFaceCount=m_spOutputVertices->count();

			m_primVerts.resize(1);
			m_primVerts[0].resize(addPointCount);

			Array<I32>& rVerts=m_primVerts[0];

			for(I32 vertexIndex=0;vertexIndex<addPointCount;vertexIndex++)
			{
				rVerts[vertexIndex]=existingPointCount+vertexIndex;
			}

			m_spOutputVertices->append(m_primVerts);

			m_spOutputProperties->set(existingFaceCount,e_openCurve,TRUE);

			if(m_radiusArray.size())
			{
				for(I32 vertexIndex=0;vertexIndex<addPointCount;vertexIndex++)
				{
					const I32 pointIndex=existingPointCount+vertexIndex;
					m_spOutputRadius->set(pointIndex,
							m_radiusArray[vertexIndex]);
				}
			}
		}
	}
	else if(Alembic::AbcGeom::IXform::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			feLog("TRANSFORM\n");
		}
#endif
		Alembic::AbcGeom::IXform ixform(a_iObj,
				Alembic::AbcGeom::kWrapExisting);

		const I32 sampleCount=ixform.getSchema().getNumSamples();

		Alembic::AbcGeom::TimeSamplingPtr sampler=
				ixform.getSchema().getTimeSampling();

		if(m_mode==e_scan)
		{
			m_xformCount++;

			if(sampleCount>1)
			{
				const Alembic::AbcGeom::chrono_t time0=
						sampler->getSampleTime(0);
				const Alembic::AbcGeom::chrono_t time1=
						sampler->getSampleTime(1);
				const Real step=time1-time0;

				if(step>1e-3 && (m_minStep<0.0 || m_minStep>step))
				{
					m_minStep=step;
				}
			}
		}

#if FALSE // FE_SAA_DEBUG
		for(I32 sampleIndex=0;sampleIndex<sampleCount;sampleIndex++)
		{
			const Alembic::AbcGeom::chrono_t time=
					sampler->getSampleTime(sampleIndex);
			feLog("sample %d/%d frame %.6G time %.6G\n",
					sampleIndex,sampleCount,time*m_fps,time);
		}
#endif

		if(m_mode==e_joint)
		{
			std::pair<Alembic::AbcGeom::index_t,Alembic::AbcGeom::chrono_t>
					nearIndex=sampler->getNearIndex(m_frame/m_fps,sampleCount);

			I32 sampleIndex=nearIndex.first;
#if FE_SAA_DEBUG
			feLog("sampleIndex %d/%d near %.6G\n",
					sampleIndex,sampleCount,m_frame/m_fps);
#endif
			FEASSERT(sampleIndex>=0);
			FEASSERT(!sampleCount || sampleIndex<sampleCount);

			Alembic::AbcGeom::XformSample sample;
			ixform.getSchema().get(sample,
					Alembic::Abc::ISampleSelector(
					Alembic::Abc::index_t(sampleIndex)));

			Alembic::AbcGeom::M44d m44d=sample.getMatrix();
			double* pRaw=m44d.getValue();

			SpatialTransform xformRelative;
			set(xformRelative,pRaw);

			const Color color(0.0,0.0,1.0);			//* TODO
			const Color localScale(1.0,1.0,1.0);	//* TODO

			const String name=path.basename();
			const String parentName=path.pathname().basename();

			setJoint(name,parentName,xformRelative,localScale,
					m_refWorld,color);
		}
	}
	else if(Alembic::AbcGeom::IFaceSet::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			feLog("FACE SET\n");
		}
#endif
	}
	else if(Alembic::AbcGeom::IXform::matches(a_iObj.getHeader()))
	{
#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			feLog("UNRECOGNIZED\n");
		}
#endif
	}

	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		Alembic::AbcGeom::IObject iChild(a_iObj,
				a_iObj.getChildHeader(childIndex).getName());

#if FE_SAA_DEBUG
		if(m_mode!=e_scan)
		{
			feLog("\nin \"%s\" child %d/%d \"%s\"\n",
					path.c_str(),childIndex,childCount,
					iChild.getName().c_str());
		}
#endif

		visit(iChild);
	}
}

void SurfaceAccessibleAbc::visitProperties(
	Alembic::AbcGeom::ICompoundProperty a_parent,
	String &a_rIndent,String a_parentPath,String a_path)
{
#if FE_SAA_DEBUG
	String oldIndent=a_rIndent;
#endif

	for(U32 propIndex=0;propIndex<a_parent.getNumProperties();propIndex++)
	{
		Alembic::AbcGeom::PropertyHeader header=
				a_parent.getPropertyHeader(propIndex);

		if(header.isCompound())
		{
			visitProperty(Alembic::AbcGeom::ICompoundProperty(a_parent,
					header.getName()),a_rIndent,a_parentPath,a_path);
		}
		else if(header.isScalar())
		{
			visitProperty(Alembic::AbcGeom::IScalarProperty(a_parent,
					header.getName()),a_rIndent,a_parentPath,a_path);
		}
		else
		{
			FEASSERT(header.isArray());

			Alembic::AbcGeom::GeometryScope geometryScope=
					Alembic::AbcGeom::kUnknownScope;

#if FE_SAA_DEBUG
			String typeName="????";
#endif

			if (Alembic::AbcGeom::IV3dGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IV3dGeomParam";
#endif

				Alembic::AbcGeom::IV3dGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IV3fGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IV3fGeomParam";
#endif

				Alembic::AbcGeom::IV3fGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IV2fGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IV2fGeomParam";
#endif

				Alembic::AbcGeom::IV2fGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IP3fGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IP3fGeomParam";
#endif

				Alembic::AbcGeom::IP3fGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IN3fGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IN3fGeomParam";
#endif

				Alembic::AbcGeom::IN3fGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IC3fGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IC3fGeomParam";
#endif

				Alembic::AbcGeom::IC3fGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IInt32GeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IInt32GeomParam";
#endif

				Alembic::AbcGeom::IInt32GeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IUInt32GeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IUInt32GeomParam";
#endif

				Alembic::AbcGeom::IUInt32GeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}
			else if (Alembic::AbcGeom::IStringGeomParam::matches(header))
			{
#if FE_SAA_DEBUG
				typeName="IStringGeomParam";
#endif

				Alembic::AbcGeom::IStringGeomParam param(
						a_parent,header.getName());
				geometryScope=param.getScope();
			}

			visitProperty(Alembic::AbcGeom::IArrayProperty(a_parent,
					header.getName()),a_rIndent,a_parentPath,a_path,
					geometryScope);

#if FE_SAA_DEBUG
			feLog("%s  datatype \"%s\" ",a_rIndent.c_str(),typeName.c_str());

			switch(geometryScope)
			{
				case Alembic::AbcGeom::kConstantScope:
					feLog("kConstantScope\n");
					break;
				case Alembic::AbcGeom::kUniformScope:
					feLog("kUniformScope\n");
					break;
				case Alembic::AbcGeom::kVaryingScope:
					feLog("kVaryingScope\n");
					break;
				case Alembic::AbcGeom::kVertexScope:
					feLog("kVertexScope\n");
					break;
				case Alembic::AbcGeom::kFacevaryingScope:
					feLog("kFacevaryingScope\n");
					break;
				case Alembic::AbcGeom::kUnknownScope:
					feLog("kUnknownScope\n");
					break;
				default:
					feLog("invalid scope\n");
			}
#endif
		}
	}

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}


void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IScalarProperty a_properties,
	String a_indent,String a_parentPath,String a_path)
{
	FEASSERT(!a_properties.isArray());

/*
	const Alembic::AbcCoreAbstract::DataType &dataType=
			a_properties.getDataType();
	const Alembic::Util::uint8_t extent=dataType.getExtent();
	Alembic::Util::Dimensions dimensions(extent);
	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr=
			Alembic::AbcCoreAbstract::AllocateArraySample(dataType,dimensions);
	const U32 sampleCount=a_properties.getNumSamples();
	for(U32 sampleIndex=0;sampleIndex<sampleCount;++sampleIndex)
	{
		a_properties.get(samplePtr->getData(),
				Alembic::Abc::ISampleSelector(
				Alembic::Abc::index_t(sampleIndex)));
		arraySize=samplePtr->size();
	};
*/

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();

	feLog("%sscalar \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),a_properties.getName().c_str(),
			interp.c_str(),dtype.str().c_str(),a_properties.getNumSamples());
#endif
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IArrayProperty a_properties,
	String a_indent,String a_parentPath,String a_path,
	Alembic::AbcGeom::GeometryScope a_geometryScope)
{
	FEASSERT(a_properties.isArray());

	const String parentName=a_properties.getParent().getName().c_str();
	const String propName=a_properties.getName().c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();
	const String dataType=dtype.str().c_str();

	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr;
	const I32 sampleCount = a_properties.getNumSamples();

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	feLog("%sarray \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),propName.c_str(),
			interp.c_str(),dataType.c_str(),sampleCount);
#endif

	Alembic::AbcGeom::TimeSamplingPtr sampler=
			a_properties.getTimeSampling();
	std::pair<Alembic::AbcGeom::index_t,Alembic::AbcGeom::chrono_t>
			nearIndex=sampler->getNearIndex(m_frame/m_fps,sampleCount);

	I32 sampleIndex=nearIndex.first;
#if FE_SAA_DEBUG
	feLog("%s  sampleIndex %d/%d near %.6G\n",
			a_indent.c_str(),sampleIndex,sampleCount,m_frame/m_fps);
#endif
	FEASSERT(sampleIndex>=0);
	FEASSERT(!sampleCount || sampleIndex<sampleCount);

	a_properties.get(samplePtr,
			Alembic::Abc::ISampleSelector(
			Alembic::Abc::index_t(sampleIndex)));

	const U32 arrayCount=samplePtr->size();
#if FE_SAA_DEBUG
	feLog("%s  arrayCount %d\n",a_indent.c_str(),arrayCount);
#endif
	if(!arrayCount)
	{
		return;
	}

	if(dataType=="string")
	{
		std::string* pString=(std::string*)samplePtr->getData();
		const String string=pString[0].c_str();

#if FE_SAA_VERBOSE
		for(U32 arrayIndex=0;arrayIndex<arrayCount && arrayIndex<8;arrayIndex++)
		{
			const String nextString=pString[arrayIndex].c_str();
			if(!arrayIndex || string!=nextString)
			{
				feLog("%s  %d/%d \"%s\"\n",a_indent.c_str(),
						arrayIndex,arrayCount,nextString.c_str());
			}
		}
#endif

		if(propName=="worldreference")
		{
			Real val[16];
			sscanf(string.c_str(),
					"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
					&val[0],&val[1],&val[2],&val[3],
					&val[4],&val[5],&val[6],&val[7],
					&val[8],&val[9],&val[10],&val[11],
					&val[12],&val[13],&val[14],&val[15]);

			set(m_refWorld,val);
		}
	}
	else if(dataType=="int32_t")
	{
		I32* i32Array=(I32*)samplePtr->getData();

#if FE_SAA_VERBOSE
		I32 minIndex=0;
		I32 maxIndex=0;
		I32 lastInteger=i32Array[0];

		for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
		{
			const I32 nextInteger=i32Array[arrayIndex];
			if((!arrayIndex || nextInteger!=lastInteger) && arrayIndex<8)
			{
				feLog("%s  %d/%d %d\n",a_indent.c_str(),
						arrayIndex,arrayCount,nextInteger);
			}
			lastInteger=nextInteger;

			if(!arrayIndex || minIndex>nextInteger)
			{
				minIndex=nextInteger;
			}
			if(!arrayIndex || maxIndex<nextInteger)
			{
				maxIndex=nextInteger;
			}
		}
		feLog("%s  min %d max %d\n",a_indent.c_str(),
				minIndex,maxIndex);
#endif

		if(propName==".faceCounts")
		{
			m_vertCountArray.resize(arrayCount);
			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				m_vertCountArray[arrayIndex]=i32Array[arrayIndex];
			}
		}
		if(propName==".faceIndices")
		{
			m_vertIndexArray.resize(arrayCount);
			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				m_vertIndexArray[arrayIndex]=i32Array[arrayIndex];
			}
		}
		if(propName==".faces")
		{
			const I32 firstFace=m_firstFaceMap[a_parentPath];
			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
//				feLog("face %d/%d %d+%d %s %s\n",arrayIndex,arrayCount,
//						firstFace,i32Array[arrayIndex],
//						a_parentPath.c_str(),a_path.c_str());
				m_spOutputName->set(firstFace+i32Array[arrayIndex],a_path);
			}
		}
	}
	else if(dataType=="uint32_t")
	{
		U32* u32Array=(U32*)samplePtr->getData();

#if FE_SAA_VERBOSE
		U32 minIndex=0;
		U32 maxIndex=0;
		U32 lastInteger=u32Array[0];

		for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
		{
			const U32 nextInteger=u32Array[arrayIndex];
			if((!arrayIndex || nextInteger!=lastInteger) && arrayIndex<8)
			{
				feLog("%s  %d/%d %d\n",a_indent.c_str(),
						arrayIndex,arrayCount,nextInteger);
			}
			lastInteger=nextInteger;

			if(!arrayIndex || minIndex>nextInteger)
			{
				minIndex=nextInteger;
			}
			if(!arrayIndex || maxIndex<nextInteger)
			{
				maxIndex=nextInteger;
			}
		}
		feLog("%s  min %d max %d\n",a_indent.c_str(),
				minIndex,maxIndex);
#endif

		if(parentName=="uv" && propName==".indices")
		{
			m_uvLookupArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				m_uvLookupArray[arrayIndex]=u32Array[arrayIndex];
			}
		}
	}
	else if(dataType=="float32_t[2]")
	{
		F32* f32Array=(F32*)samplePtr->getData();

#if FE_SAA_VERBOSE
		for(U32 arrayIndex=0;arrayIndex<arrayCount && arrayIndex<8;arrayIndex++)
		{
			const F32* pVector=&f32Array[arrayIndex*2];
			const Vector2 value(pVector[0],pVector[1]);

			feLog("%s  %d/%d %s\n",a_indent.c_str(),
					arrayIndex,arrayCount,c_print(value));
		}
#endif

		if(parentName=="uv" && propName==".vals")
		{
			m_uvScope=a_geometryScope;

			m_uvArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const F32* pVector=&f32Array[arrayIndex*2];

				set(m_uvArray[arrayIndex],
						pVector[0],pVector[1],0.0);
			}
		}
	}
	else if(dataType=="float32_t[3]")
	{
		F32* f32Array=(F32*)samplePtr->getData();

#if FE_SAA_VERBOSE
		for(U32 arrayIndex=0;arrayIndex<arrayCount && arrayIndex<8;arrayIndex++)
		{
			const F32* pVector=&f32Array[arrayIndex*3];
			const SpatialVector value(pVector[0],pVector[1],pVector[2]);

			feLog("%s  %d/%d %s\n",a_indent.c_str(),
					arrayIndex,arrayCount,c_print(value));
		}
#endif

		if(propName=="P")
		{
			m_pointArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const F32* pVector=&f32Array[arrayIndex*3];

				set(m_pointArray[arrayIndex],
						pVector[0],pVector[1],pVector[2]);
			}
		}

		if(propName=="N" || propName=="normal")
		{
			m_normalArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const F32* pVector=&f32Array[arrayIndex*3];

				set(m_normalArray[arrayIndex],
						pVector[0],pVector[1],pVector[2]);
			}
		}

		if(propName=="Cd" || propName=="color")
		{
			m_colorScope=a_geometryScope;

			m_colorArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const F32* pVector=&f32Array[arrayIndex*3];

				set(m_colorArray[arrayIndex],
						pVector[0],pVector[1],pVector[2]);
			}
		}
	}
	else if(dataType=="float64_t[3]")
	{
		F64* f64Array=(F64*)samplePtr->getData();

#if FE_SAA_VERBOSE
		for(U32 arrayIndex=0;arrayIndex<arrayCount && arrayIndex<8;arrayIndex++)
		{
			const F64* pVector=&f64Array[arrayIndex*3];
			const SpatialVector value(pVector[0],pVector[1],pVector[2]);

			feLog("%s  %d/%d %s\n",a_indent.c_str(),
					arrayIndex,arrayCount,c_print(value));
		}
#endif

		if(propName=="xgen_Pref")
		{
			m_pointRefArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const F64* pVector=&f64Array[arrayIndex*3];

				set(m_pointRefArray[arrayIndex],
						pVector[0],pVector[1],pVector[2]);
			}
		}
	}
	else if(dataType=="float64_t")
	{
		F64* f64Array=(F64*)samplePtr->getData();

#if FE_SAA_VERBOSE
		for(U32 arrayIndex=0;arrayIndex<arrayCount && arrayIndex<8;arrayIndex++)
		{
			const F64 value=f64Array[arrayIndex];

			feLog("%s  %d/%d %.6G\n",a_indent.c_str(),
					arrayIndex,arrayCount,value);
		}
#endif

		if(propName=="radiusf")
		{
			m_radiusArray.resize(arrayCount);

			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				m_radiusArray[arrayIndex]=f64Array[arrayIndex];
			}
		}
	}
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::ICompoundProperty a_properties,
	String &a_rIndent,String a_parentPath,String a_path)
{
#if FE_SAA_DEBUG
	String schema=a_properties.getMetaData().get("schema").c_str();

	feLog("%scompound \"%s\" schema '%s'\n",
			a_rIndent.c_str(),a_properties.getName().c_str(),schema.c_str());

	String oldIndent=a_rIndent;
	a_rIndent+="  ";
#endif

	visitProperties(a_properties,a_rIndent,a_parentPath,a_path);

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}

} // namespace
