/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SVO_DEBUG			FALSE
#define FE_SVO_THREAD_DEBUG		FALSE
#define FE_SVO_RECORDER			FALSE

#include "oplab/oplab.pmh"

using namespace fe;

SurfaceViewerOp::SurfaceViewerOp(void):
	m_jobs(0),
	m_lastFrame(-1.0),
	m_slack(0.0),
	m_slackInc(1.0),
	m_slackMin(0.0)
{
#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::SurfaceViewerOp\n");
#endif
}

Real SurfaceViewerOp::updateSlack(void)
{
	const Real current=m_slack;

	const Real slackGain=0.1;

	m_slack+=m_slackInc*slackGain;
	if(m_slack<m_slackMin)
	{
		m_slack=m_slackMin;
		m_slackInc=fabs(m_slackInc);
		m_slackMin=0.0;
	}
	else if(m_slack>1.0)
	{
		m_slack=1.0;
		m_slackInc=1.0;
		m_slackMin=0.0;
	}

	return current;
}

SurfaceViewerOp::~SurfaceViewerOp(void)
{
#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::~SurfaceViewerOp\n");
#endif
}

void SurfaceViewerOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<SpatialVector>("vertical")=SpatialVector(0.0,1.0,0.0);

	catalog<String>("drawStyle")="solid";
	catalog<String>("drawStyle","choice:0")="pointCloud";
	catalog<String>("drawStyle","label:0")="Point Cloud";
	catalog<String>("drawStyle","choice:1")="wireframe";
	catalog<String>("drawStyle","label:1")="Wireframe";
	catalog<String>("drawStyle","choice:2")="outline";
	catalog<String>("drawStyle","label:2")="Outline";
	catalog<String>("drawStyle","choice:3")="solid";
	catalog<String>("drawStyle","label:3")="Solid";
	catalog<String>("drawStyle","choice:4")="edgedSolid";
	catalog<String>("drawStyle","label:4")="Edged Solid";
	catalog<String>("drawStyle","choice:5")="foreshadow";
	catalog<String>("drawStyle","label:5")="Foreshadow";

	catalog<Real>("lineWidth")=1.0;
	catalog<Real>("lineWidth","high")=8.0;
	catalog<String>("lineWidth","label")="Line Width";

	catalog<bool>("finished")=FALSE;
	catalog<String>("finished","IO")="output";

	catalog<String>("keys");
	catalog<String>("keys","IO")="output";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Input Surface 2");
	catalog<bool>("Input Surface 2","optional")=TRUE;

	catalog< sp<Component> >("Input Surface 3");
	catalog<bool>("Input Surface 3","optional")=TRUE;

	catalog< sp<Component> >("Input Surface 4");
	catalog<bool>("Input Surface 4","optional")=TRUE;

	//* no auto-copy
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<bool>("Output Surface","recycle")=true;
}

void SurfaceViewerOp::post(Real a_frame)
{
#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::post %.6G jobs %d\n",a_frame,m_jobs);
#endif

	m_spGang->post(a_frame);
	m_jobs++;

	Real completed=Real(0);
	while(m_jobs)
	{
		while(m_spGang->acceptDelivery(completed))
		{
#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::post accept %.6G jobs %d\n",
					completed,m_jobs);
#endif

			m_jobs--;
		}
		milliSleep(1);
	}

#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::post done jobs %d\n",m_jobs);
#endif
}

void SurfaceViewerOp::Worker::operate(void)
{
#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::Worker::operate\n");
#endif

	BWORD bound=FALSE;
	BWORD firstFrame=TRUE;

	Real job=Real(0);
	while(TRUE)
	{
		//* lock down for each iteration
		sp<SurfaceViewerOp> spSurfaceViewerOp=m_hpSurfaceViewerOp;
		if(spSurfaceViewerOp.isNull())
		{
			break;
		}

		sp<QuickViewerI> spQuickViewer=spSurfaceViewerOp->quickViewer();
		if(spQuickViewer.isNull())
		{
			milliSleep(10);
			continue;
		}

		if(!bound)
		{
			spQuickViewer->open();

			sp<DrawI> spDrawI=spQuickViewer->getDrawI();
			if(spDrawI.isNull())
			{
				continue;
			}

			spSurfaceViewerOp->drawHandler()->bind(spDrawI);

			sp<DrawMode> spDrawMode=spDrawI->drawMode();
			if(spDrawMode.isValid())
			{
				spDrawMode->setBackfaceCulling(FALSE);
				spDrawMode->setFrontfaceCulling(FALSE);
			}

			sp<HandlerI> spHandlerI(spDrawI);
			if(spHandlerI.isValid())
			{
				spQuickViewer->insertEventHandler(spDrawI);
			}

			bound=TRUE;
		}

		sp<DrawI> spDrawI=spQuickViewer->getDrawI();
		if(spDrawI.isValid())
		{
			sp<DrawMode> spDrawMode=spDrawI->drawMode();
			if(spDrawMode.isValid())
			{
				spDrawMode->setDrawStyle(spSurfaceViewerOp->drawStyle());
				spDrawMode->setLineWidth(spSurfaceViewerOp->lineWidth());
			}
		}


#if	FE_SVO_THREAD_DEBUG
		feLog("SurfaceViewerOp::Worker::operate check\n");
#endif

#if FE_SVO_RECORDER
		m_hpJobQueue->waitForJob(job);
#else
		if(m_hpJobQueue->take(job))
#endif
		{
			if(job<0)
			{
#if	FE_SVO_THREAD_DEBUG
				feLog("SurfaceViewerOp::Worker::operate quit %.6G\n",job);
#endif

				m_hpJobQueue->deliver(job);
				break;
			}

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate do job %.6G\n",job);
#endif

			const U32 us=systemMicroseconds();
			const I32 delta=us-m_us;
			m_us=us;

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate delta %d us\n",delta);
#endif

			sp<DrawableI> spDrawable1=spSurfaceViewerOp->drawable1();
			sp<DrawableI> spDrawable2=spSurfaceViewerOp->drawable2();
			sp<DrawableI> spDrawable3=spSurfaceViewerOp->drawable3();
			sp<DrawableI> spDrawable4=spSurfaceViewerOp->drawable4();

			spSurfaceViewerOp->drawHandler()->bindDrawable1(spDrawable1);
			spSurfaceViewerOp->drawHandler()->bindDrawable2(spDrawable2);
			spSurfaceViewerOp->drawHandler()->bindDrawable3(spDrawable3);
			spSurfaceViewerOp->drawHandler()->bindDrawable4(spDrawable4);

			spSurfaceViewerOp->drawHandler()->setFrame(job);
			spSurfaceViewerOp->drawHandler()->setMicroseconds(delta);

			sp<SurfaceI> spSurface=spDrawable1;
			if(spSurface.isValid())
			{
				const Real slack=spSurfaceViewerOp->updateSlack();

				SpatialVector center;
				Real radius;
				if(!spSurfaceViewerOp->drawHandler()->getPickRegion(
						center,radius))
				{
					center=spSurface->center();
					radius=spSurface->radius();
				}

				//* NOTE tweak
				const Real minDistance=radius*
						(firstFrame? 10.0: (4.0-3.5*pow(slack,8.0)));
				const Real maxDistance=radius*
						(10.0+90.0*slack);
				const Real maxOffset=radius*
						(firstFrame? 0.0: (0.1+0.9*slack));
				const Real response=
						(firstFrame? 1.0: (1.0-0.98*slack));

				sp<ControllerI> spController=
						spQuickViewer->getControllerI();
				SpatialVector lookPoint=spController->lookPoint();
				const SpatialVector correction=center-lookPoint;
				const Real offset=magnitude(correction);
				const Real scalar=offset-maxOffset;
				if(scalar>0.0)
				{
					lookPoint+=correction*(response*scalar/offset);
					spController->setLookPoint(lookPoint);
				}

				BWORD changeDistance=FALSE;

				Real distance=spController->distance();
				if(distance<minDistance)
				{
					distance=(1.0-response)*distance+response*minDistance;
					changeDistance=TRUE;
				}
				else if(distance>maxDistance)
				{
					distance=(1.0-response)*distance+response*maxDistance;
					changeDistance=TRUE;
				}

				if(changeDistance)
				{
					const Real azimuth=spController->azimuth();
					const Real downangle=spController->downangle();

					spController->setLookAngle(azimuth,downangle,distance);
				}

//				feLog("look at %s dist %.6G\n",c_print(lookPoint),distance);

				firstFrame=FALSE;
			}

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate done job %.6G\n",job);
#endif

			m_hpJobQueue->deliver(job);

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate delivered job %.6G\n",job);
#endif
		}
#if	FE_SVO_THREAD_DEBUG
		else
		{
			feLog("SurfaceViewerOp::Worker::operate no jobs\n");
		}
#endif

		sp<QuickViewerI> spQuickViewerI=spSurfaceViewerOp->quickViewer();
		if(spQuickViewerI.isValid())
		{
			spQuickViewerI->run(1);
		}
	}

#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::Worker::operate done\n");
#endif
}

void SurfaceViewerOp::handle(Record& a_rSignal)
{
	const String drawStyle=catalog<String>("drawStyle");

	m_drawStyle=DrawMode::e_solid;
	if(drawStyle=="pointCloud")
	{
		m_drawStyle=DrawMode::e_pointCloud;
	}
	else if(drawStyle=="wireframe")
	{
		m_drawStyle=DrawMode::e_wireframe;
	}
	else if(drawStyle=="outline")
	{
		m_drawStyle=DrawMode::e_outline;
	}
	else if(drawStyle=="edgedSolid")
	{
		m_drawStyle=DrawMode::e_edgedSolid;
	}
	else if(drawStyle=="foreshadow")
	{
		m_drawStyle=DrawMode::e_foreshadow;
	}

	m_lineWidth=catalog<Real>("lineWidth");

	SpatialVector vertical=unitSafe(catalog<SpatialVector>("vertical"));
	if(magnitude(vertical)<0.999)
	{
		set(vertical,0.0,1.0,0.0);
	}

#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::handle\n");
#endif

	if(m_spGang.isNull())
	{
		m_spGang=new Gang<Worker,Real>();
		m_spGang->start(sp<Counted>(this),1);
	}

	//* see surface/test/xSurface.cc

	const Real frame=currentFrame(a_rSignal);
	if(m_lastFrame!=frame)
	{
		m_spDrawable1=NULL;
		m_spDrawable2=NULL;
		m_spDrawable3=NULL;
		m_spDrawable4=NULL;
		m_lastFrame=frame;
	}

	for(I32 surfaceIndex=0;surfaceIndex<4;surfaceIndex++)
	{
		sp<DrawableI>& rspDrawable=
				surfaceIndex==0? m_spDrawable1:
				(surfaceIndex==1? m_spDrawable2:
				(surfaceIndex==2? m_spDrawable3: m_spDrawable4));

		if(rspDrawable.isNull())
		{
			String attrName="Input Surface";
			if(surfaceIndex)
			{
				attrName.sPrintf("Input Surface %d",surfaceIndex+1);
			}

			sp<SurfaceAccessibleI> spSurfaceAccessible;
			access(spSurfaceAccessible,attrName,e_quiet);

			sp<SurfaceI> spInput;
			access(spInput,attrName,surfaceIndex? e_quiet: e_warning);

			if(!surfaceIndex && spInput.isNull())
			{
				return;
			}

			if(spInput.isValid() && spInput->radius()<1e-3)
			{
				catalog<String>("warning")+=
						attrName + " has no apparent content;";
			}

			if(spSurfaceAccessible.isValid() && spInput.isValid())
			{
				spInput->setName(spSurfaceAccessible->name());
			}

			rspDrawable=spInput;
		}
	}

	if(m_spDrawable1.isNull())
	{
		catalog<String>("error")="input surface not drawable";
		return;
	}

	if(m_spDrawHandler.isNull())
	{
		m_spDrawHandler=
				Library::create<DrawHandler>("SurfaceViewerOp::DrawHandler");
	}

	if(m_spEventHandler.isNull())
	{
		m_spEventHandler=
				Library::create<EventHandler>("SurfaceViewerOp::EventHandler");
	}

#if FE_SVO_RECORDER
	//* TODO param
	if(m_spVideoRecorder.isNull())
	{
		m_spVideoRecorder=registry()->create("*.VideoRecorder");
		m_spVideoRecorder->configure("test/chainvideo",240);
	}
#endif

	if(m_spQuickViewerI.isNull())
	{
		registry()->master()->catalog()->catalog<String>("hint:DrawI")=
				"*.DrawCached";

		m_spQuickViewerI=registry()->create("QuickViewerI");
		if(!m_spQuickViewerI.isValid())
		{
			catalog<String>("error")="couldn't create viewer";
			return;
		}

		const Color darkBlue(0.20,0.20,0.25);
		m_spQuickViewerI->getWindowI()->setBackground(darkBlue);

		sp<ControllerI> spController=m_spQuickViewerI->getControllerI();
		spController->setVertical(vertical);
		spController->setLookPoint(SpatialVector(0.0,0.0,0.0));
		spController->setLookAngle(-80.0,30.0,1.0f);

		if(m_spDrawHandler.isValid())
		{
			m_spDrawHandler->setSurfaceViewerOp(sp<SurfaceViewerOp>(this));
			m_spQuickViewerI->insertDrawHandler(m_spDrawHandler);
		}

		if(m_spEventHandler.isValid())
		{
			m_spEventHandler->setSurfaceViewerOp(sp<SurfaceViewerOp>(this));
			m_spQuickViewerI->insertEventHandler(m_spEventHandler);
		}

		if(m_spVideoRecorder.isValid())
		{
			feLog("SurfaceViewerOp::handle adding VideoRecorder\n");

			m_spQuickViewerI->insertDrawHandler(m_spVideoRecorder);
		}
	}

	BWORD closed=FALSE;

	if(m_spQuickViewerI.isValid())
	{
		sp<ControllerI> spController=m_spQuickViewerI->getControllerI();
		closed=spController->closeEvent();
	}

	post(closed? -1: frame);

	catalog<bool>("finished")=closed;

#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::handle done\n");
#endif
}

SurfaceViewerOp::DrawHandler::DrawHandler(void):
	m_frame(0.0),
	m_us(0),
	m_pickRegion(FALSE),
	m_keepEdges(FALSE),
	m_impactSurfaceIndex(-1)
{
	m_spDrawWire=new DrawMode();
	m_spDrawWire->setDrawStyle(DrawMode::e_wireframe);
	m_spDrawWire->setPointSize(6.0);
	m_spDrawWire->setLineWidth(4.0);
	m_spDrawWire->setAntialias(TRUE);

	m_spDrawImpact=new DrawMode();
	m_spDrawImpact->setDrawStyle(DrawMode::e_foreshadow);
	m_spDrawImpact->setLineWidth(2.0);
	m_spDrawImpact->setLit(FALSE);
	m_spDrawImpact->setAntialias(TRUE);

	m_spDrawOutline=new DrawMode();
	m_spDrawOutline->copy(m_spDrawImpact);
	m_spDrawOutline->setLineWidth(6.0);
}

void SurfaceViewerOp::DrawHandler::setFrame(Real a_frame)
{
	if(m_frame!=a_frame)
	{
		m_frame=a_frame;
		m_keepEdges=FALSE;
		m_spSurfaceImpact=NULL;

		if(m_impactSurfaceIndex<0)
		{
			m_spImpactSurface=NULL;
			m_spImpactAccessible=NULL;
		}
		else
		{
			m_hpSurfaceViewerOp->access(m_spImpactAccessible,
					m_impactSurfaceName,e_quiet);

			m_spImpactSurface=NULL;
			if(m_impactSurfaceIndex==0)
			{
				m_spImpactSurface=m_spDrawable1;
			}
			else if(m_impactSurfaceIndex==1)
			{
				m_spImpactSurface=m_spDrawable2;
			}
			else if(m_impactSurfaceIndex==2)
			{
				m_spImpactSurface=m_spDrawable3;
			}
			else if(m_impactSurfaceIndex==3)
			{
				m_spImpactSurface=m_spDrawable4;
			}
		}
	}
}

BWORD SurfaceViewerOp::DrawHandler::getPickRegion(
	SpatialVector& a_rCenter,Real& a_rRadius) const
{
	if(!m_pickRegion)
	{
		return FALSE;
	}
	a_rCenter=m_pickCenter;
	a_rRadius=m_pickRadius;
	return TRUE;
}

void SurfaceViewerOp::DrawHandler::unpick(void)
{

	if(m_spSurfaceImpact.isNull() && m_pickName.empty())
	{
		m_spImpactAccessible=NULL;
		m_impactSurfaceName="";
		m_impactSurfaceIndex= -1;
		m_spImpactSurface=NULL;
	}

	m_spSurfaceImpact=NULL;
	m_pickName="";
	m_keepEdges=FALSE;
}

void SurfaceViewerOp::DrawHandler::handle(Record& a_render)
{
	if(!m_asViewer.scope().isValid())
	{
		m_asViewer.bind(a_render.layout()->scope());
	}
	if(!m_asViewer.viewer_perspective.check(a_render) ||
			!m_asViewer.viewer_layer.check(a_render) ||
			!m_asViewer.viewer_layer(a_render))
	{
		return;
	}

	const I32& rLayer=m_asViewer.viewer_layer(a_render);
	if(rLayer==1)
	{
		m_pickRegion=FALSE;

		//* main render

		const Color lightbrown(0.8,0.7,0.5);

		if(m_spDrawable1.isValid())
		{
			m_spDrawable1->draw(m_spDrawI,
					m_spDrawable1==m_spImpactSurface? &lightbrown: NULL);
		}

		if(m_spDrawable2.isValid())
		{
			m_spDrawable2->draw(m_spDrawI,
					m_spDrawable2==m_spImpactSurface? &lightbrown: NULL);
		}

		if(m_spDrawable3.isValid())
		{
			m_spDrawable3->draw(m_spDrawI,
					m_spDrawable3==m_spImpactSurface? &lightbrown: NULL);
		}

		if(m_spDrawable4.isValid())
		{
			m_spDrawable4->draw(m_spDrawI,
					m_spDrawable4==m_spImpactSurface? &lightbrown: NULL);
		}

		sp<SurfaceViewerOp> spSurfaceViewerOp=m_hpSurfaceViewerOp;

		const I32 primitiveIndex=m_spSurfaceImpact.isValid()?
				m_spSurfaceImpact->primitiveIndex(): -1;

		if(spSurfaceViewerOp.isValid() && m_spImpactAccessible.isValid())
		{
			String partName;

			if(primitiveIndex<0)
			{
				partName=m_pickName;
			}
			else
			{
				m_partAttr="";

				Array<SurfaceAccessibleI::Spec> specs;

				m_spImpactAccessible->attributeSpecs(specs,
						SurfaceAccessibleI::e_primitive);

				const I32 specCount=specs.size();
				for(I32 specIndex=0;specIndex<specCount;specIndex++)
				{
					const SurfaceAccessibleI::Spec& rSpec=specs[specIndex];

					const String attrName=rSpec.name();
					if(attrName!="name" && attrName!="part")
					{
						continue;
					}

					const String attrType=rSpec.typeName();
					if(attrType!="string")
					{
						continue;
					}

					String valueString;

					sp<SurfaceAccessorI> spSurfaceAccessorI;
					spSurfaceViewerOp->access(spSurfaceAccessorI,
							m_spImpactAccessible,
							OperatorSurfaceCommon::e_primitive,
							attrName);

					if(spSurfaceAccessorI.isValid())
					{
						m_partAttr=attrName;
						partName=spSurfaceAccessorI->string(primitiveIndex);

						//* priority
						if(attrName=="name")
						{
							break;
						}
					}
				}
			}

			sp<SurfaceAccessorI> spPrimitiveVertices;
			spSurfaceViewerOp->access(spPrimitiveVertices,
					m_spImpactAccessible,
					OperatorSurfaceCommon::e_primitive,
					OperatorSurfaceCommon::e_vertices);

			sp<SurfaceAccessorI> spPrimitiveProperties;
			spSurfaceViewerOp->access(spPrimitiveProperties,
					m_spImpactAccessible,
					OperatorSurfaceCommon::e_primitive,
					OperatorSurfaceCommon::e_properties);

			const I32 primCount=spPrimitiveVertices->count();

			I32 curveCount=0;

			if(!partName.empty())
			{
				if(partName!=m_pickName)
				{
					spSurfaceViewerOp->setSlack(1.0,1.0,0.5);
					m_keepEdges=FALSE;
				}

				FEASSERT(!m_partAttr.empty());

				SpatialVector pickSum;
				SpatialVector pickMin;
				SpatialVector pickMax;
				I32 pickCount=0;

				m_spDrawI->pushDrawMode(m_spDrawWire);

				sp<SurfaceAccessorI> spPrimitivePart;
				spSurfaceViewerOp->access(spPrimitivePart,
						m_spImpactAccessible,
						OperatorSurfaceCommon::e_primitive,
						m_partAttr);

				const Color green(0.0,1.0,0.0,0.2);
				const Color hardGreen(0.0,1.0,0.0,0.6);
				const Color hardYellow(1.0,1.0,0.0,0.8);

				for(I32 primIndex=0;primIndex<primCount;primIndex++)
				{
					const String onePart=spPrimitivePart->string(primIndex);
					if(onePart==partName)
					{
						const BWORD openCurve=
								(spPrimitiveProperties.isValid() &&
								spPrimitiveProperties->integer(
								primIndex,e_openCurve));
						curveCount+=openCurve;

						const I32 subCount=
								spPrimitiveVertices->subCount(primIndex);

						const I32 lineCount=(subCount+!openCurve);
						SpatialVector line[lineCount];

						for(I32 subIndex=0;subIndex<subCount;subIndex++)
						{
							const SpatialVector point=
									spPrimitiveVertices->spatialVector(
									primIndex,subIndex);

							line[subIndex]=point;

							if(!pickCount)
							{
								pickSum=point;
								pickMin=point;
								pickMax=point;
							}
							else
							{
								pickSum+=point;
								for(I32 m=0;m<3;m++)
								{
									if(pickMin[m]>point[m])
									{
										pickMin[m]=point[m];
									}
									else if(pickMax[m]<point[m])
									{
										pickMax[m]=point[m];
									}
								}
							}
							pickCount++;
						}

						if(!openCurve)
						{
							line[lineCount-1]=line[0];
						}

						m_spDrawI->drawLines(line,NULL,lineCount,
								DrawI::e_strip,false,
								primIndex==primitiveIndex? &hardGreen: &green);

						if(primIndex==primitiveIndex)
						{
							m_spDrawI->drawPoints(line,NULL,subCount,
									FALSE,&hardYellow);
						}
					}
				}

				m_spDrawI->popDrawMode();

				if(pickCount)
				{
					m_pickRegion=TRUE;
					m_pickCenter=pickSum/Real(pickCount);
					m_pickRadius=magnitude(pickMin-m_pickCenter);
					const Real otherRadius=
							magnitude(pickMax-m_pickCenter);
					if(m_pickRadius<otherRadius)
					{
						m_pickRadius=otherRadius;
					}
				}
			}
			else
			{
				for(I32 primIndex=0;primIndex<primCount;primIndex++)
				{
					const BWORD openCurve=
							(spPrimitiveProperties.isValid() &&
							spPrimitiveProperties->integer(
							primIndex,e_openCurve));
					curveCount+=openCurve;
				}
			}

			if(m_pickName.empty() && m_spSurfaceImpact.isNull())
			{
				m_keepEdges=FALSE;
			}
			else if(!curveCount)
			{
				if(!m_keepEdges)
				{
					createEdges(m_spImpactAccessible,m_partAttr,m_edgeMap);
					m_keepEdges=TRUE;
				}

				sp<ViewI> spView=m_spDrawI->view();
				sp<ViewMode> spViewMode=spView->viewMode();
				SpatialTransform camera=spViewMode->camera();
				SpatialTransform invCamera;
				invert(invCamera,camera);

//				const SpatialVector cameraDir= -invCamera.column(2);
				const SpatialVector cameraPos=invCamera.column(3);

				const Color alphaBlack(0.0,0.0,0.0,0.3);

				m_spDrawI->pushDrawMode(m_spDrawOutline);

				drawOutline(m_spDrawI,m_edgeMap,partName,cameraPos,alphaBlack);

				m_spDrawI->popDrawMode();
			}

			m_pickName=partName;
		}
		else
		{
			m_pickName="";
			m_keepEdges=FALSE;
		}

		if(!m_keepEdges)
		{
			m_edgeMap.clear();
		}

		sp<DrawableI> spDrawableImpacted=m_spImpactSurface;
		if(spDrawableImpacted.isValid())
		{
			sp<SurfaceTriangles> spSurfaceTriangles=spDrawableImpacted;

			//* NOTE trianglated surfaces only (lots of hair can be slow)
			if(spSurfaceTriangles.isValid())
			{
				const Color blue(0.0,0.5,1.0,0.1);

				m_spDrawI->pushDrawMode(m_spDrawWire);
				spDrawableImpacted->draw(m_spDrawI,&blue);
				m_spDrawI->popDrawMode();
			}
		}

		sp<DrawableI> spDrawableImpact=m_spSurfaceImpact;
		if(spDrawableImpact.isValid())
		{
			const Color alphaYellow(1.0,1.0,0.0,0.2);

			m_spDrawI->pushDrawMode(m_spDrawImpact);
			spDrawableImpact->draw(m_spDrawI,&alphaYellow);
			m_spDrawI->popDrawMode();
		}

		return;
	}

	if(rLayer==2)
	{
		const Color lightBlue(0.5,0.5,1.0);
		const Color yellow(1.0,1.0,0.0);
		const Color grey(0.6,0.6,0.6);

		SpatialVector location(10.0,70.0,0.0);

		m_spDrawI->drawAlignedText(location,"CTRL camera",lightBlue);
		location[1]-=10.0;

		m_spDrawI->drawAlignedText(location,"ESC  deselect",lightBlue);
		location[1]-=10.0;

		m_spDrawI->drawAlignedText(location,"f    reframe",lightBlue);

		set(location,10.0,30.0,0.0);

		String text;
		if(m_us>0)
		{
			text.sPrintf("%.1f ms",0.001*m_us);
			m_spDrawI->drawAlignedText(location,text,yellow);
		}
		location[1]-=10.0;

		text.sPrintf("Frame %.6G",m_frame);
		m_spDrawI->drawAlignedText(location,text,yellow);

		sp<ViewI> spViewI=m_spDrawI->view();
		const I32 winy=height(spViewI->viewMode()->viewport())-16;
		set(location,10.0,winy,0.0);

		if(m_spImpactSurface.isValid() && m_spSurfaceImpact.isValid())
		{
			m_spDrawI->drawAlignedText(location,
					m_spImpactSurface->name(),yellow);
			location[1]-=10.0;

			text.sPrintf("pick   %s",
					c_print(m_spSurfaceImpact->intersection()));
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			text.sPrintf("normal %s",c_print(m_spSurfaceImpact->normal()));
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			const Barycenter<Real> bary=m_spSurfaceImpact->barycenter();
			text.sPrintf("bary   %s face %d",c_print(bary),
					m_spSurfaceImpact->face());
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			text.sPrintf("uv     %s",c_print(m_spSurfaceImpact->uv()));
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			text.sPrintf("du     %s",c_print(m_spSurfaceImpact->du()));
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			text.sPrintf("dv     %s",c_print(m_spSurfaceImpact->dv()));
			m_spDrawI->drawAlignedText(location,text,yellow);
			location[1]-=10.0;

			I32 pointIndex0= -1;
			I32 pointIndex1= -1;
			I32 pointIndex2= -1;
			I32 nearestPointIndex= -1;
			I32 nearestSubIndex=0;

			sp<SurfaceSearchable::Impact> spSurfaceSearchableImpact=
					m_spSurfaceImpact;
			if(spSurfaceSearchableImpact.isValid())
			{
				pointIndex0=spSurfaceSearchableImpact->pointIndex0();
				pointIndex1=spSurfaceSearchableImpact->pointIndex1();

				sp<SurfaceTriangles::Impact> spSurfaceTrianglesImpact=
						m_spSurfaceImpact;
				if(spSurfaceTrianglesImpact.isValid())
				{
					pointIndex2=spSurfaceTrianglesImpact->pointIndex2();

					text.sPrintf("points %d %d %d",
							pointIndex0,pointIndex1,pointIndex2);
					m_spDrawI->drawAlignedText(location,text,yellow);
					location[1]-=10.0;

					const Real bary2=Real(1)-bary[0]-bary[1];
					nearestPointIndex=bary[0]>bary[1]?
						(bary[0]>bary2? pointIndex0: pointIndex2):
						(bary[1]>bary2? pointIndex1: pointIndex2);

					nearestSubIndex=bary[0]>bary[1]?
						(bary[0]>bary2? 0: 2):
						(bary[1]>bary2? 1: 2);
				}
				else
				{
					text.sPrintf("points %d %d",
							pointIndex0,pointIndex1);
					m_spDrawI->drawAlignedText(location,text,yellow);
					location[1]-=10.0;

					nearestPointIndex=pointIndex0;
				}

#if FALSE
				const I32 partitionIndex=
						m_spSurfaceImpact->partitionIndex();

				sp<SurfaceSearchable> spSurfaceSearchable=m_spImpactSurface;
				const String partitionName=spSurfaceSearchable.isValid()?
						m_spImpactSurface->partitionName(partitionIndex): "";

				text.sPrintf("part %d %s",partitionIndex,partitionName.c_str());
				m_spDrawI->drawAlignedText(location,text,yellow);
				location[1]-=10.0;
#endif
			}

			const I32 primitiveIndex=m_spSurfaceImpact->primitiveIndex();
			const I32 triangleIndex=m_spSurfaceImpact->triangleIndex();

			const I32 primitiveCount=m_spImpactAccessible.isValid()?
					m_spImpactAccessible->count(
					SurfaceAccessibleI::e_primitive): 0;

			sp<SurfaceViewerOp> spSurfaceViewerOp=m_hpSurfaceViewerOp;

			if(spSurfaceViewerOp.isValid() && m_spImpactAccessible.isValid())
			{
				Array<SurfaceAccessibleI::Spec> specs;

				for(I32 pass=0;pass<4;pass++)
				{
					SurfaceAccessibleI::Element element;
					I32 elementIndex=0;
					I32 subIndex=0;

					if(pass==0)
					{
						if(nearestPointIndex<0)
						{
							continue;
						}

						elementIndex = nearestPointIndex;
						element = SurfaceAccessibleI::e_point;
					}
					else if(pass==1)
					{
						elementIndex = primitiveIndex;
						subIndex = nearestSubIndex;
						element = SurfaceAccessibleI::e_vertex;
					}
					else if(pass==2)
					{
						elementIndex = primitiveIndex;
						element = SurfaceAccessibleI::e_primitive;
					}
					else if(pass==3)
					{
						element = SurfaceAccessibleI::e_detail;
					}

					m_spImpactAccessible->attributeSpecs(specs,element);

					const I32 specCount=specs.size();
					if(!specCount)
					{
						continue;
					}

					if(pass==0)
					{
						location[1]-=10.0;

						const I32 pointCount=m_spImpactAccessible.isValid()?
								m_spImpactAccessible->count(
								SurfaceAccessibleI::e_point): 0;

						text.sPrintf("point     %-14d",nearestPointIndex);
						if(pointCount>0)
						{
							text.sPrintf("%s of %d",text.c_str(),pointCount);
						}

						m_spDrawI->drawAlignedText(location,text,yellow);
						location[1]-=10.0;
					}
					else if(pass==1)
					{
						text.sPrintf("vertex %d",subIndex);
						m_spDrawI->drawAlignedText(location,text,yellow);
						location[1]-=10.0;
					}
					else if(pass==2)
					{
						text.sPrintf("primitive %-14d",primitiveIndex);
						if(primitiveCount>0)
						{
							text.sPrintf("%s of %d",
									text.c_str(),primitiveCount);
						}
						if(primitiveIndex!=triangleIndex)
						{
							text.sPrintf("%s (triangle %d)",
									text.c_str(),triangleIndex);
						}
						m_spDrawI->drawAlignedText(location,text,yellow);
						location[1]-=10.0;

						sp<SurfaceAccessorI> spSurfaceAccessorI;
						spSurfaceViewerOp->access(spSurfaceAccessorI,
								m_spImpactAccessible,
								OperatorSurfaceCommon::e_primitive,
								OperatorSurfaceCommon::e_vertices);
						if(spSurfaceAccessorI.isValid())
						{
							const I32 subCount=
									spSurfaceAccessorI->subCount(primitiveIndex);

							text.sPrintf("%-9s %-14s %d",
									"integer","subCount",subCount);
							m_spDrawI->drawAlignedText(location,text,yellow);
							location[1]-=10.0;
						}
					}
					else if(pass==3)
					{
						text.sPrintf("detail");
						m_spDrawI->drawAlignedText(location,text,yellow);
						location[1]-=10.0;
					}

					for(I32 specIndex=0;specIndex<specCount;specIndex++)
					{
						const SurfaceAccessibleI::Spec& rSpec=specs[specIndex];
						const String attrName=rSpec.name();
						const String attrType=rSpec.typeName();

						if(attrName=="surf:points" || attrName=="surf:vertices")
						{
							continue;
						}

						String valueString;

						sp<SurfaceAccessorI> spSurfaceAccessorI;
						spSurfaceViewerOp->access(spSurfaceAccessorI,
								m_spImpactAccessible,
								OperatorSurfaceCommon::Element(element),
								attrName);

						if(spSurfaceAccessorI.isValid())
						{
							if(attrType=="string")
							{
								valueString=spSurfaceAccessorI->string(
										elementIndex,subIndex);
							}
							else if(attrType=="integer")
							{
								const I32 value=spSurfaceAccessorI->integer(
										elementIndex,subIndex);
								valueString.sPrintf("%d",value);
							}
							else if(attrType=="real")
							{
								const Real value=spSurfaceAccessorI->real(
										elementIndex,subIndex);
								valueString.sPrintf("%.6G",value);
							}
							else if(attrType=="vector3" || attrType=="color")
							{
								const SpatialVector value=
										spSurfaceAccessorI->spatialVector(
										elementIndex,subIndex);
								valueString.sPrintf("%s",c_print(value));
							}
						}

						text.sPrintf("%-9s %-14s %s",
								attrType.c_str(),attrName.c_str(),
								valueString.c_str());
						m_spDrawI->drawAlignedText(location,text,yellow);
						location[1]-=10.0;
					}

					location[1]-=10.0;
				}
			}
		}
		else
		{
			for(I32 surfaceIndex=0;surfaceIndex<4;surfaceIndex++)
			{
				sp<SurfaceI> spSurface;
				if(surfaceIndex==0)
				{
					spSurface=m_spDrawable1;
				}
				else if(surfaceIndex==1)
				{
					spSurface=m_spDrawable2;
				}
				else if(surfaceIndex==2)
				{
					spSurface=m_spDrawable3;
				}
				else if(surfaceIndex==3)
				{
					spSurface=m_spDrawable4;
				}
				if(spSurface.isNull())
				{
					continue;
				}

				m_spDrawI->drawAlignedText(location,spSurface->name(),
						surfaceIndex==m_impactSurfaceIndex? yellow: grey);
				location[1]-=10.0;
			}
		}

		return;
	}
}

void SurfaceViewerOp::EventHandler::handle(Record& a_rSignal)
{
	m_event.bind(a_rSignal);

	if(m_event.isIdleMouse() || m_event.isPoll())
	{
		return;
	}

	sp<SurfaceViewerOp> spSurfaceViewerOp=m_hpSurfaceViewerOp;

	if(spSurfaceViewerOp.isValid())
	{
		spSurfaceViewerOp->react(m_event);
	}
}

void SurfaceViewerOp::react(WindowEvent& a_rEvent)
{
//	feLog("SurfaceViewerOp::react %s\n",print(a_rEvent).c_str());

	if(a_rEvent.isKeyPress(WindowEvent::e_keyEscape))
	{
		m_spDrawHandler->unpick();

		return;
	}

	U32 item=a_rEvent.item();
	BWORD shifted=FALSE;
	if(item&WindowEvent::e_keyShift)
	{
		shifted=TRUE;
		item^=WindowEvent::e_keyShift;
	}
	BWORD ctrled=FALSE;
	if(item&WindowEvent::e_keyControl)
	{
		ctrled=TRUE;
		item^=WindowEvent::e_keyControl;
	}

	if(shifted || ctrled)
	{
		return;
	}

	if(a_rEvent.isKeyPress(WindowEvent::Item('f')))
	{
		setSlack(1.0,-1.0,0.0);
	}
	else if(a_rEvent.isKeyPress(WindowEvent::e_keyCursorLeft))
	{
		catalog<String>("keys")+="h";
	}
	else if(a_rEvent.isKeyPress(WindowEvent::e_keyCursorRight))
	{
		catalog<String>("keys")+="l";
	}
	else if(a_rEvent.isKeyPress(WindowEvent::e_itemAny) &&
		item<WindowEvent::e_keyAsciiMask)
	{
		String character;
		character.sPrintf("%c",item);

		catalog<String>("keys")+=character;
	}

	if(!a_rEvent.isMousePress(WindowEvent::e_itemLeft))
	{
		return;
	}

	I32 nearestIndex= -1;
	sp<SurfaceI> spNearestSurface;
	sp<SurfaceI::ImpactI> spNearestImpact;

	I32 mousex=0;
	I32 mousey=0;
	a_rEvent.getMousePosition(&mousex,&mousey);

	sp<DrawI> spDrawI=m_spQuickViewerI->getDrawI();
	sp<ViewI> spViewI=spDrawI->view();

	const I32 winy=height(spViewI->viewMode()->viewport())-1-mousey;

	const SpatialVector rayOrigin=
			spViewI->unproject(mousex,winy,0.0,ViewMode::e_perspective);
	const SpatialVector target=
			spViewI->unproject(mousex,winy,1.0,ViewMode::e_perspective);
	const SpatialVector rayDirection=unitSafe(target-rayOrigin);

//	feLog("pick origin %s dir %s\n",c_print(rayOrigin),c_print(rayDirection));

	for(I32 surfaceIndex=0;surfaceIndex<4;surfaceIndex++)
	{
		sp<SurfaceI> spSurface;
		if(surfaceIndex==0)
		{
			spSurface=m_spDrawable1;
		}
		else if(surfaceIndex==1)
		{
			spSurface=m_spDrawable2;
		}
		else if(surfaceIndex==2)
		{
			spSurface=m_spDrawable3;
		}
		else if(surfaceIndex==3)
		{
			spSurface=m_spDrawable4;
		}
		if(spSurface.isNull())
		{
			continue;
		}

		try
		{
			const Real maxDistance= -1.0;
			sp<SurfaceI::ImpactI> spImpact=spSurface->rayImpact(
					rayOrigin,rayDirection,maxDistance);
			if(spImpact.isValid())
			{
				if(spNearestImpact.isNull() ||
						spNearestImpact->distance()>spImpact->distance())
				{
					nearestIndex=surfaceIndex;
					spNearestSurface=spSurface;
					spNearestImpact=spImpact;
				}
			}
		}
		catch(...)
		{
		}
	}

	if(m_spDrawHandler.isValid())
	{
		String attrName="Input Surface";
		if(nearestIndex)
		{
			attrName.sPrintf("Input Surface %d",nearestIndex+1);
		}

		sp<SurfaceAccessibleI> spNearestAccessible;
		access(spNearestAccessible,attrName,e_quiet);

		m_spDrawHandler->setSurfaceImpact(
				spNearestSurface==m_spDrawHandler->impactSurface()?
				spNearestImpact:
				sp<SurfaceI::ImpactI>(NULL));

		m_spDrawHandler->setImpactAccessible(spNearestAccessible);
		m_spDrawHandler->setImpactSurface(spNearestSurface);
		m_spDrawHandler->setImpactSurfaceName(attrName);
		m_spDrawHandler->setImpactSurfaceIndex(nearestIndex);
	}
}
