/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "core.pmh"

namespace fe
{

TypeMaster::TypeMaster(void):
	m_name("TypeMaster")
{
}

TypeMaster::~TypeMaster(void)
{
}

sp<BaseType> TypeMaster::lookupName(const String &name)
{
	SAFEGUARD;
	string_type_map::iterator it = m_hNameMap.find(name);
	sp<BaseType> spType;
	if(it != m_hNameMap.end())
	{
		spType = it->second;
	}

	return spType;
}

String TypeMaster::lookupBaseType(const sp<BaseType>& type)
{
	SAFEGUARD;
	type_string_map::iterator it = m_hTypeMap.find(type);
	String spName;
	if(it != m_hTypeMap.end())
	{
		spName = it->second;
	}

	return spName;
}

sp<BaseType> TypeMaster::lookupType(const TypeInfo &typeInfo)
{
	SAFEGUARD;

/*
	feLog("find %s\n",typeInfo.ref().name());
	for(	type_info_map::iterator t_it = m_hTypeidMap.begin();
			t_it != m_hTypeidMap.end();
			t_it++)
	{
		feLog("%d %s\n", t_it->first==typeInfo,
				t_it->first.ref().name());
	}
*/
	type_info_map::iterator it = m_hTypeidMap.find(typeInfo);
	if(it != m_hTypeidMap.end())
	{
		return it->second;
	}
	else
	{
		feLog("find [%s]\n",typeInfo.ref().name());
		for(	type_info_map::iterator t_it = m_hTypeidMap.begin();
				t_it != m_hTypeidMap.end();
				t_it++)
		{
			feLog("%d [%s] %s\n", t_it->first==typeInfo,
					t_it->first.ref().name(), typeInfo.ref().name());
		}
		feLog("TypeMaster::lookupType not found: %s\n", typeInfo.ref().name());
		sp<BaseType> spType;
		return spType;
	}
}

void TypeMaster::peek(Peeker &peeker)
{
	SAFEGUARD;
	peeker.nl();
	peeker.cat("known types:");
	peeker.in();
	for(	type_info_map::iterator t_it = m_hTypeidMap.begin();
			t_it != m_hTypeidMap.end();
			t_it++)
	{
		peeker.nl();
		peeker.str().cat("%-30s", t_it->first.ref().name());
		String tnames;
		for(	string_type_map::iterator n_it = m_hNameMap.begin();
				n_it != m_hNameMap.end();
				n_it++)
		{
			if(n_it->second == t_it->second)
			{
				peeker.bold();
				peeker.str().cat(" %s", n_it->first.c_str());
				peeker.normal();
			}
		}
	}
	peeker.out();

	peeker.nl();
	peeker.cat("conversions:");
	peeker.in();
	for(	in_out_converter_map::iterator io_it = m_hConverter.begin();
			io_it != m_hConverter.end();
			io_it++)
	{
		for(	out_converter_map::iterator o_it = io_it->second.begin();
				o_it != io_it->second.end();
				o_it++)
		{
			String sIn, sOut;
			for(	type_info_map::iterator t_it = m_hTypeidMap.begin();
					t_it != m_hTypeidMap.end();
					t_it++)
			{
				if(t_it->second == io_it->first)
				{	sIn = t_it->first.ref().name(); }
				if(t_it->second == o_it->first)
				{	sOut = t_it->first.ref().name(); }
			}
			peeker.nl();
			peeker.str().cat("in:%-10s out:%-10s", sIn.c_str(), sOut.c_str());
			if(o_it->second.isValid())
			{
			}
			else
			{
				peeker.bold();
				peeker.str().cat(" invalid");
				peeker.normal();
			}
		}
	}
	peeker.out();
}


void TypeMaster::registerConversion(sp<TypeConversion> spConverter,
		sp<BaseType> spInType, sp<BaseType> spOutType)
{
	SAFEGUARD;
	sp<TypeConversion> existing=lookupConversion(spInType,spOutType);
	if(existing.isValid())
	{
		return;
	}

	m_hConverter[spInType][spOutType] = spConverter;
}


sp<TypeConversion> TypeMaster::lookupConversion(sp<BaseType> spInType,
		sp<BaseType> spOutType)
{
	SAFEGUARD;
	return m_hConverter[spInType][spOutType];
}

void TypeMaster::reverseLookup(sp<BaseType> basetype, std::list<String> &names)
{
	SAFEGUARD;
	for(string_type_map::iterator it = m_hNameMap.begin();
			it != m_hNameMap.end(); it++)
	{
		if(it->second == basetype)
		{
			names.push_back(it->first);
		}
	}
}

String TypeMaster::reverseLookup(sp<BaseType> basetype)
{
	String namelist;
	std::list<String> names;
	reverseLookup(basetype, names);
	for(std::list<String>::iterator i_name = names.begin();
		i_name != names.end(); i_name++)
	{
		if(i_name != names.begin())
		{
			namelist.cat(" ");
		}
		namelist.cat(*i_name);
	}
	return namelist;
}

sp<BaseType> TypeMaster::assertTypeInternal(const TypeInfo &typeInfo,
		sp<BaseType> spT, const String &name)
{
	SAFEGUARD;
	type_info_map::iterator it = m_hTypeidMap.find(typeInfo);
	// did not find type
	if(it == m_hTypeidMap.end())
	{
		// but name already used
		if(this->lookupName(name).isValid())
		{
			feX("fe::TypeMaster::assertType",
				"type name mismatch on %s", name.c_str());
		}
		m_hTypeidMap[typeInfo] = spT;
		m_hNameMap[name] = spT;
		m_hTypeMap[spT] = name;
		return spT;
	}
	// found type
	else
	{
		sp<BaseType> spType = it->second;
		// and the name exists
		sp<BaseType> namedType = this->lookupName(name);
		if(namedType.isValid())
		{
			// but is not the same type (should now be impossible)
			if(namedType != spType)
			{
				feX("fe::TypeMaster::assertType",
					"(impossibility) type name mismatch on %s", name.c_str());
			}
		}
		// name does not exist; add it
		else
		{
			m_hNameMap[name] = spType;
			m_hTypeMap[spT] = name;
		}
		return spType;
	}
}

}
